// PenTest.cpp
// This file is public domain software (PDS).
// Copyright (C) 2017 Katayama Hirofumi MZ <katayama.hirofumi.mz@gmail.com>
#include "MWindowBase.hpp"

class PenTest : public MWindowBase
{
public:
    PenTest()
    {
    }

    virtual LRESULT CALLBACK
    WindowProcDx(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
    {
        switch (uMsg)
        {
            HANDLE_MSG(hwnd, WM_CREATE, OnCreate);
            HANDLE_MSG(hwnd, WM_DESTROY, OnDestroy);
            HANDLE_MSG(hwnd, WM_PAINT, OnPaint);
        }
        return DefaultProcDx();
    }

    BOOL OnCreate(HWND hwnd, LPCREATESTRUCT lpCreateStruct)
    {
        return TRUE;
    }

    virtual void ModifyWndClassDx(WNDCLASSEX& wcx)
    {
        wcx.hbrBackground = CreateSolidBrush(RGB(255, 255, 0));
    }

    void OnDestroy(HWND hwnd)
    {
        PostQuitMessage(0);
    }

    void OnPaint(HWND hwnd)
    {
        PAINTSTRUCT ps;
        HDC hDC = BeginPaint(hwnd, &ps);
        if (hDC)
        {
            HENHMETAFILE hEMF = NULL;
            HDC hdcEMF = CreateEnhMetaFile(hDC, NULL, NULL, TEXT("This is a test"));
            if (hdcEMF)
            {
                LOGBRUSH lb;
                GetObject(GetStockObject(BLACK_BRUSH), sizeof(lb), &lb);
                HPEN hPen = ExtCreatePen(PS_GEOMETRIC | PS_SOLID | PS_ENDCAP_ROUND | PS_JOIN_ROUND, 10, &lb, 0, NULL);
                HGDIOBJ hPenOld = SelectObject(hdcEMF, hPen);
                MoveToEx(hdcEMF, 0, 0, NULL);
                LineTo(hdcEMF, 100, 100);
                SelectObject(hdcEMF, hPenOld);
                hEMF = CloseEnhMetaFile(hdcEMF);
            }
            if (hEMF)
            {
                RECT rc;
                SetRect(&rc, 0, 0, 100, 100);
                PlayEnhMetaFile(hDC, hEMF, &rc);
            }
            EndPaint(hwnd, &ps);
        }
    }

    BOOL StartDx(LPSTR lpCmdLine, INT nCmdShow)
    {
        if (!CreateWindowDx(NULL, TEXT("PenTest"), WS_OVERLAPPEDWINDOW, 0,
                            CW_USEDEFAULT, CW_USEDEFAULT, 350, 280))
        {
            MsgBoxDx(TEXT("CreateWindowDx failed"), NULL, MB_ICONERROR);
            return FALSE;
        }

        ShowWindow(m_hwnd, nCmdShow);
        UpdateWindow(m_hwnd);
        return TRUE;
    }

    INT_PTR RunDx()
    {
        MSG msg;
        while (GetMessage(&msg, NULL, 0, 0))
        {
            TranslateMessage(&msg);
            DispatchMessage(&msg);
        }
        return INT(msg.wParam);
    }
};

INT WINAPI
WinMain(HINSTANCE   hInstance,
        HINSTANCE   hPrevInstance,
        LPSTR       lpCmdLine,
        INT nCmdShow)
{
    PenTest app;
    if (app.StartDx(lpCmdLine, nCmdShow))
    {
        return app.RunDx();
    }
    return -1;
}
